import numpy as np 
import pandas as pd
import datetime
import matplotlib.pyplot as plt 
import seaborn as sns
import os 
wd = os.getcwd()

def write_training_log(model, config, result):
    
    os.makedirs(f'./log/{model.__arch__}', exist_ok= True)
    date = str(datetime.datetime.now())[:-7]
    columns = ['Date','Description', 'Version', 'Optimizer', 'Device', '#Epochs', 'Learning_rate', 'Batch_size'] \
                        + ['t_train', 't_train_epoch', 't_data'] \
                        + ['Train_loss', 'Train_accuracy', 'Val_loss', 'Val_acc', 'Test_acc', 'Test_f1', 'Test_precision', 'Test_recall'] \
                        

    # Write scoring results in a .csv file
    session_result_path = f'./log/{model.__arch__}/session_result.csv'
    if os.path.exists(session_result_path):
        f = pd.read_csv(session_result_path, index_col = [0])
    else:
        f = pd.DataFrame(columns=columns)
    
    
    infos = [date, config.exp_name + '_' + config.method, model.__version__, config.optimizer, config.device, config.n_epochs, config.lr, config.batch_size]
    t = [result['training_time'], np.mean(result['per_epoch_training_time']), result['data_processing_time']]
    train_loss = result['training_loss'][-1]
    train_acc = result['training_accuracy'][-1]
    val_loss = result['validation_loss'][-1]
    val_acc = result['validation_accuracy'][-1]
    results = [train_loss, train_acc, val_loss, val_acc] + list(result['test_score'].values())

    test_label = list(result['test_class_accuracy'].keys())
    class_acc_cols = [f'{label}_acc' for label in test_label]
    columns += class_acc_cols
    class_acc = list(result['test_class_accuracy'].values())

    # print([infos + t + results + class_acc])
    # print(columns)
    f = pd.concat([f, pd.DataFrame([infos + t + results + class_acc],columns = columns)])
    f.columns = columns
    f.to_csv(session_result_path)

    # Write training log in a .txt file
    with open(f'./log/{model.__arch__}/session_log.txt','a') as f:

        f.writelines([
                    f'======================================================================================\n',
                    f'Date: {date} | Description: {config.exp_name} | Model version: {model.__version__}\n',
                    f'Optimizer: {config.optimizer} | Device: {config.device} | Epochs: {config.n_epochs} | Learning rate: {config.lr} | Batch size: {config.batch_size}\n',
                    f"Early stopping: {result['early_stopping_epoch']}\n",
                    f"Training loss: {' '.join([str(num) for num in np.round(result['training_loss'],2)])}\n",
                    f"Training accuracy: {' '.join([str(num) for num in np.round(result['training_accuracy'],2)])}\n",
                    f"Validation loss: {' '.join([str(num) for num in np.round(result['validation_loss'],2)])}\n",
                    f"Validation accuracy: {' '.join([str(num) for num in np.round(result['validation_accuracy'],2)])}\n",
                    f"Flatten cf: {' '.join([str(num) for num in result['test_confusion_matrix'].flatten()])}\n",
                    ])  


def plot_training_result(model, config, result, savefig = True):
    plt.rcParams.update({'font.size': 12})
    # Learning curves
    train_loss = result['training_loss']
    train_acc = result['training_accuracy']
    val_loss = result['validation_loss']
    val_acc = result['validation_accuracy']

    # test scores
    test_score = result['test_score']
    scores = list(result['test_score'].keys())
    
    # Confusion matrices
    cf = result['test_confusion_matrix']
    
    f, ax = plt.subplots(1,3,figsize = (14,5))
    ax[0].plot(train_loss,'r',label = 'Tr.loss')
    ax[0].plot(train_acc,'r--',label = 'Tr.acc')
    ax[0].plot(val_loss,'b', label = 'Val.loss')
    ax[0].plot(val_acc,'b--', label = 'Val.acc')
    ax[0].set_xlabel('Epoch')
    ax[0].set_title('Accuracy curve and loss curve.')
    ax[0].grid()
    ax[0].legend()

    w = 0.3
    h = [test_score[k] for k in scores]
    plt.bar
    ax[1].bar(np.arange(len(scores)), h, width = w)
    ax[1].set_xticks(np.arange(len(scores)),scores)
    ax[1].set_title('Test accuracy and weighted scores')
    ax[1].set_ylim(0,100)
    
    test_label = result['test_class_accuracy'].keys()
    sns.heatmap(np.round(cf*100,2), ax = ax[2], annot= True, cbar = False, xticklabels=test_label, yticklabels =test_label)
    ax[2].set_title('Confusion matrix')     
    ax[2].set_xlabel('Predicted label')
    ax[2].set_ylabel('True label')
    plt.suptitle(config.exp_name)
    plt.tight_layout()

    if savefig == True:
        os.makedirs(f'./log/{model.__arch__}', exist_ok= True)
        
        d = str(datetime.date.today())
        p = os.path.join(f'./log/{model.__arch__}/{config.exp_name}_{model.__arch__}_{d}.png')
        plt.savefig(p)      

def format_filenames(folder):
    if os.path.exists(f'./data/{folder}'):
        dir = os.listdir(f'./data/{folder}')
        for n in dir:
            if not n.startswith(folder):
                newname = folder + '_' + n
                os.rename(f'./data/{folder}/{n}', f'./data/{folder}/{newname}')
    else:
        raise RuntimeError(f"Path './data/{folder}' does not exist.")


import os
import pandas as pd 
import glob
from easydict import EasyDict 
from copy import deepcopy as dc

import torch
from datetime import date

# ============================= Dataset =============================
datasets = {
            'zt': ['0zt','8zt','16zt'], 
            'ct': ['0ct','8ct','16ct'],
            'BCOA-Wheat': ['BCOA-Wheat'],
            'Cannabis-hemp': ['non-viruliferous-hemp', 'viruliferous-hemp'],
            'Cannabis-potato': ['non-viruliferous-potato', 'viruliferous-potato'],
            'ArabidopsisGPA': ['ArabidopsisGPA'],
            'wheatrnai': ['wheatrnai'], 
            'SBA-HostSurface': ['SBA-HostSurface'],
            'SBA-Rag5': ['SBA-Rag5'],
            'sorghumaphid': ['sorghumaphid'], 
            }

def get_dataset_group(group):
    if group == 1 or group == 'BCOA1': # BCOA1
        # return datasets['zt']
        return datasets['zt'] + datasets['ct']
    elif group == 2 or group == 'BCOA2': # BCOA2
        return datasets['BCOA-Wheat']  
    elif group == 3 or group == 'CA': # Cannabis aphid
        return datasets['Cannabis-hemp'] + datasets['Cannabis-potato']  
    elif group == 4 or group == 'GPA': # GPA
        return datasets['ArabidopsisGPA']
    elif group == 5 or group == 'SA': # Soybean
        return datasets['SBA-Rag5']
        # return datasets['SBA-Rag5'] + datasets['SBA-HostSurface']
    elif group == 6 or group == 'SorgA': # Sorghum
        return datasets['sorghumaphid']
    elif group == 7 or group == 'RNAi': # Wheat RNAi
        return datasets['wheatrnai']
    elif group == 99 or group == 'combined':
        all_names = []
        for i in range(1,6):
            all_names += get_dataset_group(i)
        return all_names
    else:
        try:
            return datasets[group]
        except:
            raise ValueError('Unsupported value.')

# ============================= Label map =============================
waveform_labels = ['NP', 'C', 'E1', 'E2', 'F', 'G', 'pd']
ana_labels = [1, 2, 4, 5, 6, 7, 8]
encoded_labels = [0, 1, 2, 3, 4, 5, 6]

def create_map(list1, list2):
    map = {}
    for i in range(len(list1)):
        map[list1[i]] = list2[i]
    return map

# ============================= Format =============================
def format_data(data, label, arch = None, seq_length = None):
    label = pd.Series(label).map(create_map(ana_labels, encoded_labels)).to_numpy()
    return EasyDict({'data':data,'label':label})

# ============================= Read input =============================
import numpy as np 
import os 
wd = os.getcwd()

def get_filename(name, data_path = '../data'):
    filenames = os.listdir(f'{data_path}/{name}_ANA')
    unique = []
    for name in filenames:
        name = name[:-4]
        unique.append(name)
    return unique
    
def read_signal(filename: str, data_path = '../data') -> tuple:
    '''
        Input: 
            filename: name of the recording
        Output:
            tuple of data signal and analysis dataframe 
    '''
    # Read signals
    d = []
    dataset_name = filename.split('_')[0]
    dir = os.listdir(f'{data_path}/{dataset_name}')
    rec_components = glob.glob(f'{data_path}/{dataset_name}/{filename}.*')

    # Read the signal (.A0x) files
    for file_path in rec_components:
        x = pd.read_csv(file_path, low_memory = False, delimiter=";", header = None, usecols=[1])
        d.append(x)
    data = pd.concat(d)
    data = data.to_numpy().squeeze(1)

    # Read the analysis (.ANA) files
    try:
        ana_path = f'{data_path}/{dataset_name}_ANA/{filename}.ANA'
        ana = pd.read_csv(ana_path, encoding='utf-16', delimiter = '\t',header = None, usecols=[0,1])
        ana.columns = ['label','time']
        ana = ana[(ana['label'] != 9) & (ana['label'] != 10) & (ana['label'] != 11)]
        ana.drop_duplicates(subset='time',inplace=True)
        ana.index = [i for i in range(len(ana))]
    except Exception as e:
        print(e)
        ana = None 
    return data, ana

def get_index(ana):
    '''
    Input.
        ana: analysis file
    Output.
        index: dictionary containing intervals of all wave types found in the analysis file 
            1 ~ np, 2 ~ c, 4 ~ e1, 5 ~ e2, 6 ~ f, 7 ~ g, 8 ~ pd
    '''
    index = {}
    n = len(ana)
    for i in range(0,n-1):
        start, end = ana.loc[i:i+1,'time'].tolist()
        if ana.loc[i,'label'] == 1:
            try:
                index['NP'].append([start,end])
            except:
                index['NP'] = [[start,end]]
        elif ana.loc[i,'label'] == 2:
            try:
                index['C'].append([start,end])
            except:
                index['C'] = [[start,end]]
        elif ana.loc[i,'label'] == 3:
            try:
                index['E1e'].append([start,end])
            except:
                index['E1e'] = [[start,end]]
        elif ana.loc[i,'label'] == 4:
            try:
                index['E1'].append([start,end])
            except:
                index['E1'] = [[start,end]]
        elif ana.loc[i,'label'] == 5:
            try:
                index['E2'].append([start,end])
            except:
                index['E2'] = [[start,end]]
        elif ana.loc[i,'label'] == 6:
            try:
                index['F'].append([start,end])
            except:
                index['F'] = [[start,end]]
        elif ana.loc[i,'label'] == 7:
            try:
                index['G'].append([start,end])
            except:
                index['G'] = [[start,end]]
        elif ana.loc[i,'label'] == 8:
            try:
                index['pd'].append([start,end])
            except:
                index['pd'] = [[start,end]]
        elif ana.loc[i,'label'] == 9:
            try:
                index['pd-S-II-2'].append([start,end])
            except:
                index['pd-S-II-2'] = [[start,end]]
        elif ana.loc[i,'label'] == 10:
            try:
                index['pd-S-II-3'].append([start,end])
            except:
                index['pd-S-II-3'] = [[start,end]]
        elif ana.loc[i,'label'] == 11:
            try:
                index['D'].append([start,end])
            except:
                index['D'] = [[start,end]]
        # elif ana.loc[i,'label'] == 11:
        #     try:
        #         index['pd-L'].append([start,end])
        #     except:
        #         index['pd-L'] = [[start,end]]
        # elif ana.loc[i,'label'] == 13:
        #     try:
        #         index['pd-L-II-2'].append([start,end])
        #     except:
        #         index['pd-L-II-2'] = [[start,end]]
        # elif ana.loc[i,'label'] == 14:
        #     try:
        #         index['pd-L-II-2'].append([start,end])
        #     except:
        #         index['pd-L-II-2'] = [[start,end]]
        
    return index 

def extract_sample(wave_array,ana_file,wave_type,which):
    '''
        Extract one waveform sample from the whole signal
    '''
    ana = dc(ana_file)
    ana['time'] = ana['time'].apply(lambda x: x*100)
    ana['time'] = ana['time'].astype(int)
    wave_indices = get_index(ana)
    start,end = wave_indices[wave_type][which]
    return wave_array[start:end]




import torch 
import torch.nn as nn 
from .NN import *

import numpy as np 
import pandas as pd 
import matplotlib.pyplot as plt 

import time
import datetime 
import os 
from tqdm import tqdm 

from .Dataset import EPGDataset
from ..utils import utils, metrics, visualization
from ..dataset_utils import datagenerator, dataloader, datahelper



device = 'cuda' if torch.cuda.is_available() else 'cpu'


def get_model(config):

    if config.method == 'raw':
        input_size = 1024
        in_channels = 1
    elif config.method == 'fft':
        input_size = 513
        in_channels = 1
    elif config.method == 'wavelet': 
        input_size = 515
        in_channels = 4
    if config.arch in ['mlp', 'cnn1d', 'resnet']:
        assert config.method in ['raw', 'fft', 'wavelet'], f'Unexpected feature of type {config.method} for model of type {config.arch}'
    if config.arch == 'mlp':
        return MLP(input_size = input_size)
    elif config.arch == 'cnn1d':
        return CNN1D(input_size = input_size)
    elif config.arch == 'resnet':
        return ResNet(input_size = input_size)
    elif config.arch == 'cnn2d':
        assert config.method in ['gaf', 'spectrogram', 'scalogram'], f'Unexpected feature of type {config.method} for model of type {config.arch}'
        if config.method == 'gaf':
            input_size = 64
        elif config.method == 'spectrogram':
            input_size = 65
        elif config.method == 'scalogram':
            input_size = 65
        return CNN2D(input_size = input_size)

class EPGSegment:
    def __init__(self, config, inference = False):
        self.config = config

        # Dataset
        if inference == False:
            print('Training mode.')
            self.data_path = config.data_path 
            self.dataset_name = config.dataset_name
            self.dataset = EPGDataset(self.data_path, self.dataset_name)
        else:
            print('Inference mode, skip loading data.')
        
        # Model/ optimizers
        self.device = device
        self.model = get_model(self.config).to(self.device)        
        self.lr = config.lr 
        self.batch_size = config.batch_size
        if config.optimizer == 'Adam':
            self.optimizer = torch.optim.Adam(self.model.parameters(), lr = config.lr)
        elif config.optimizer == 'SGD':
            self.optimizer = torch.optim.SGD(self.model.parameters(), lr = config.lr, momentum= 0.9, weight_decay= 0.001)   
        else:
            raise ValueError('Unsupported optimizer.')
        self.loss_fn = nn.NLLLoss().to(self.device)
        self.n_epochs = config.n_epochs

        # Configs for inputs
        self.window_size = config.window_size
        self.hop_length = config.hop_length
        self.scope = config.scope
        self.method = config.method 
        self.scale = config.scale 

        # Environment 
        self.random_state = 28
        self._is_model_trained = False 
        self._is_dataloaders_available = False

        # Result
        self.train_result_ = {  'training_loss': [], 
                                'training_accuracy': [], 
                                'validation_loss': [],
                                'validation_accuracy': [], 
                                'test_class_accuracy': [], 
                                'test_score': [], 
                                'test_confusion_matrix': [],
                                'training_time' :0, 
                                'data_processing_time':0, 
                                'per_epoch_training_time': []}

    def get_dataloaders(self, r = [0.7, 0.2, 0.1]):
        print('Obtaining dataloders ...')
        self.dataset.generate_sliding_windows(  window_size = self.window_size, 
                                                hop_length  = self.hop_length, 
                                                method      = self.config.method,
                                                scale       = self.scale,
                                                verbose     = True)
        data, labels = self.dataset.windows, self.dataset.labels
        if 'cnn' in self.model.__type__ :
            if self.config.method != 'wavelet':
                unsqueeze = True
        self.train_loader, self.val_loader, self.test_loader = dataloader.get_loaders(data, labels, 
                                                                                        r = [0.7, 0.2, 0.1], 
                                                                                        batch_size=self.batch_size, 
                                                                                        unsqueeze = unsqueeze,
                                                                                        random_state = self.random_state)
        self._is_dataloaders_available = True 
        
    #######################################
    ######### TRAINING ML MODELS ##########
    #######################################
    def train_epoch(self):
        self.model.train()
        t0 = time.perf_counter()
        training_loss = 0; n_correct = 0; n_samples = 0

        for x_batch,y_batch in self.train_loader:    
            
            x_batch, y_batch = x_batch.to(self.device), y_batch.to(self.device)
            output = self.model(x_batch)
            predicted = torch.argmax(output.detach(),-1)
            n_samples += len(y_batch.ravel())
            n_correct += (predicted == y_batch).sum().item()     
            
            loss = self.loss_fn(output,y_batch.ravel())
            training_loss += loss.item()
            self.optimizer.zero_grad()
            loss.backward()
            # _ = torch.nn.utils.clip_grad_norm_(self.model.parameters(), 0.1)
            self.optimizer.step()
            
        training_loss = training_loss/(len(self.train_loader))
        accuracy = n_correct/n_samples
        self.train_result_['per_epoch_training_time'].append(time.perf_counter()-t0)
        self.train_result_['training_loss'].append(training_loss)
        self.train_result_['training_accuracy'].append(accuracy)
        return training_loss, accuracy 

    def evaluate(self, task, verbose = True):
        # For Task 1: Classifying an input segment of fixed size (default = 1024)
        self.model.eval()
        if task == 'validate':
            loader = self.val_loader 
        elif task == 'test': 
            loader = self.test_loader
        else:
            raise ValueError("task is either 'validate' or 'test'.")

        mean_loss = 0
        with torch.no_grad():
            pred_windows = []
            true_label = []
            t0 = time.perf_counter()
            for x_batch,y_batch in loader:
                x_batch, y_batch = x_batch.to(self.device), y_batch.to(self.device)

                output = self.model(x_batch)
                loss = self.loss_fn(output,y_batch.ravel())
                mean_loss += loss.item()

                predicted = torch.argmax(output, -1)
                pred_windows.extend(predicted.ravel().cpu().numpy())
                true_label.extend(y_batch.ravel().cpu().numpy())

        if task == 'validate': 
            mean_loss = mean_loss/len(loader)
            self.train_result_['validation_loss'].append(mean_loss)
            accuracy = np.mean(np.array(true_label) == np.array(pred_windows))
            self.train_result_['validation_accuracy'].append(accuracy)
            return mean_loss, accuracy
        elif task == 'test':
            results = metrics.scoring(true_label, pred_windows)
            c = results['confusion_matrix']
            scores = results['scores']
            self.train_result_['test_score'] = scores
            self.train_result_['test_confusion_matrix'] = c
            test_label = set(true_label)
            id_to_label = {v: k for k, v in self.dataset.label_to_id.items()}
            label_to_name = {v: k for k, v in self.dataset.name_to_label.items()}
            self.train_result_['test_class_accuracy'] = {label_to_name[id_to_label[id]]: np.round(c[i,i]*100,2) for i,id in enumerate(test_label)}
            self.pred_windows = pred_windows
            if verbose == True:
                print(f'Accuracy: {scores["accuracy"]}, Average f1: {scores["f1"]}') 
                print(f'Class accuracy: {self.train_result_["test_class_accuracy"]}')
                print('Finished testing!')

    def train(self, early_stop = True, patience = 5, min_delta = 0.01, verbose = True):
        # Initialization 
        if self._is_dataloaders_available == False:
            self.get_dataloaders()
        early_stopper = EarlyStopper(patience=patience, min_delta = min_delta) if early_stop == True else None
        print('Training...') if verbose == True else None
        self.train_result_['early_stopping_epoch'] = self.n_epochs
        _is_early_stopped = False

        # Training loop
        t0 = time.perf_counter()
        for epoch in tqdm(range(self.n_epochs), desc = 'Training'):
            
            tr_loss, tr_acc = self.train_epoch()
            val_loss, val_acc = self.evaluate(task = 'validate')

            if verbose == True:
                if (epoch %10 == 0) or (epoch == self.n_epochs - 1):
                    print(f"Epoch [{epoch+1}/{self.n_epochs}] | Train loss: {tr_loss:.4f} | Val. loss: {val_loss:.4f} | Train acc: {tr_acc:.4f} | Val. acc: {val_acc:.4f}") 

            if early_stop == True:
                if early_stopper.early_stop(self.train_result_['validation_loss'][-1]):   
                    if _is_early_stopped == False:   
                        print(f'Early stopping occured at epoch {epoch+1+patience} after {patience} epochs of changes less than {min_delta} in validation accuracy. Validation loss: {self.train_result_["validation_loss"][-1]:.4f}')       
                        # Test and save the checkpoint at early stopped epochs
                        self.evaluate(task = 'test')
                        self.save_checkpoint(f'early_stopped_{epoch}')
                        self.train_result_['early_stopping_epoch'] = epoch   
                        _is_early_stopped = True 

        self.train_result_['training_time'] = time.perf_counter() - t0       
        self._is_model_trained = True
        print('Finished training!') if verbose == True else None

    #######################################
    ########## EPG SEGMENTATION ###########
    #######################################

    def segment(self, recording_name, verbose = False):
        # For Task 2: Output an analysis file (segmentation)
        if (self._is_model_trained == False):
            raise Warning('Model is not trained.')
        self.recording_name = recording_name
        if isinstance(recording_name, str):
            recData = self.dataset.loadRec(recName = recording_name)
            self.recording, self.ana = recData['recording'], recData['ana']
            print(f'File name: {recording_name}')
        elif isinstance(recording_name, int):
            dat = self.dataset[recording_name]
            self.recording, self.ana = dat['recording'], dat['ana']
            print(f'File name: {dat["name"]}')
        if self.ana is None:
            print('Ground-truth annotation was not detected.')
        test_hop_length = self.window_size//self.scope

        self.model.eval()
        
        # Initial segmentation
        print('Preparing data...') if verbose == True else None
        data = datagenerator.generate_sliding_windows_single(self.recording, self.ana, 
                                                        window_size = self.window_size,
                                                        hop_length = test_hop_length, 
                                                        method = self.method, 
                                                        scale = self.scale, 
                                                        task = 'test')

        if self.ana is not None:
            self.input, self.true_segmentation = data
        else: 
            self.input = data
        # Reshape and convert to torch.tensor
        input = torch.from_numpy(self.input).float().to(self.device)
        if self.model.__type__ == 'cnn':
            if self.config.method != 'wavelet':
                input = input.unsqueeze(1)

        # Predict each segment
        print('Generating segmentation ...') if verbose == True else None
        out = []
        for i in range(input.shape[0]):
            out.append(self.model(input[i:(i+1)]).detach().cpu().numpy())
        
        self.log_pred_proba = np.array(out)
        pred_windows = np.argmax(out, axis=-1).ravel()

        # Generate segmentation
        pred_segmentation = []
        for i in range(len(pred_windows)):
            pred_segmentation.extend([pred_windows[i]]*test_hop_length)
        pred_segmentation = np.array(pred_segmentation)
        pred_segmentation = extend(pred_segmentation, self.true_segmentation)

        # Scoring
        if self.ana is not None: 
            self.overlap_rate = np.round(np.mean(self.true_segmentation == pred_segmentation)*100, 2)
            print(f'Overlap rate: {self.overlap_rate}') if verbose == True else None

        # map to ground_truth labels  
        self.pred_segmentation = pd.Series(pred_segmentation).map({0: 1, 1: 2, 2: 4, 3: 5, 4: 6, 5: 7, 6: 8}).to_numpy() 
        self.pred_ana = to_ana(self.pred_segmentation) 
        self.pred_ana = self.pred_ana[['label','time']]
        return self.pred_ana 

    #################################################
    ########## PLOT/SAVE RESULT UTILITIES ###########
    #################################################

    def save_analysis(self, name: str = ''):

        os.makedirs('./prediction/ANA', exist_ok = True)
        dir = os.listdir('./prediction/ANA')
        index = len(dir) + 1
        if name == '':
            save_name = f'./prediction/ANA/Untitled_{index}.ANA'
        else:
            save_name = f'./prediction/ANA/{name}.ANA'
        self.pred_ana.to_csv(save_name,sep = '\t',header = None,index=None)
        print(f'Analysis saved to {save_name}')      
    def write_train_log(self):
        utils.write_training_log(self.model, self.config, self.train_result_)
                            
    def plot_train_result(self, savefig = False):
        utils.plot_training_result(self.model, self.config, self.train_result_, savefig)
        
    def save_checkpoint(self, name = ''):
        date = str(datetime.date.today())
        os.makedirs(f'./checkpoints/{self.model.__arch__}', exist_ok = True)

        if name == '':
            saved_name = f'arch-{self.model.__arch__}.version-{self.model.__version__}.window-{self.config.window_size}.'
            saved_name += f'method-{self.config.method}.scale-{self.config.scale}.optimizer-{self.config.optimizer}.'
            saved_name += f'epochs-{self.config.n_epochs}.lr-{self.config.lr}.batchsize-{self.config.batch_size}'
            save_path = f'./checkpoints/{self.model.__arch__}/{saved_name}.{date}.{self.config.exp_name}.json'
        else:
            save_path = f'./checkpoints/{self.model.__arch__}/{name}'
        torch.save(self.model, save_path)
        print(f'Parameters saved to {save_path}.')

    def plot_segmentation(self, which = 'pred_vs_gt', hour = None, range = None, savefig = False, name: str = ''): 
        visualization.plot_gt_vs_pred_segmentation(self.recording, self.ana, self.pred_ana, hour, range, which, savefig)
                
    def plot_interactive(self, which = 'prediction', smoothen = False):
        if which == 'prediction':
            visualization.interactive_visualization(self.recording, self.pred_ana, smoothen, title = which)
        elif which == 'ground_truth':
           visualization.interactive_visualization(self.recording, self.ana, smoothen, title = which)
        else:
            raise ValueError("Param which must be either 'prediction' or 'ground_truth'.")

    def reset(self):
        self._is_model_trained = False 
        self.train_result_ = {'training_loss': [], 'training_accuracy': [], 'validation_loss': [], 'validation_accuracy': [], 
                       'test_class_accuracy': [], 'test_score': [], 'test_confusion_matrix': [],
                       'training_time' :0, 'data_processing_time':0, 'per_epoch_training_time': []}
        self.model = get_model(self.config).to(self.device)       

    def load_checkpoint(self, path):
        try:
            self.model = torch.load(f'./checkpoints/{path}')
        except:
            self.model = torch.load(f'./checkpoints/{self.config.arch}/{path}')
        self._is_model_trained = True
        print('Checkpoint loaded.')
  
def to_ana(segmentation):
    # create *.ANA file
    ana_time = [0]
    ana_label = []
    tmp_label = segmentation[0]
    for i in range(1, len(segmentation)):
        if segmentation[i] != segmentation[i-1]: 
            ana_label.append(tmp_label)
            ana_time.append(i/100)
            tmp_label = segmentation[i]
    ana_label += [segmentation[i], 99]
    ana_time += [(len(segmentation) + 1)/100]
    pred_ana = pd.DataFrame({'time': ana_time, 'label': ana_label})
    
    return pred_ana

def repeat(array, coef):
    shape = array.shape 
    if len(shape) == 3:
        if shape[1] == 1:
            tmp = array.reshape(([shape[0],shape[2]]))
    n = len(array)
    repeated = []
    for i in range(n):
        repeated.extend([array[i]] * coef)
    return np.array(repeated)

def extend(arr1, target):
    '''
        Repeat the last axis of arr1 until 
        as long as target array (or reach target length)
    '''
    if isinstance(target, np.ndarray):
        ext_len = len(target) - len(arr1)
    else: 
        ext_len = target - len(arr1)
    if ext_len < 0:
        raise RuntimeError('Cannot repeat. Target length is less than input length')
    if len(arr1.shape) == 1: 
        extension = np.repeat(arr1[-1], ext_len)
    else:
        extension = np.tile(arr1[-1,:], (ext_len,1))
    return np.concatenate([arr1, extension])

class EarlyStopper:
    def __init__(self, patience=1, min_delta=0):
        self.patience = patience
        self.min_delta = min_delta
        self.counter = 0
        self.min_validation_loss = float('inf')

    def early_stop(self, validation_loss):
        if np.abs(validation_loss - self.min_validation_loss) > self.min_delta:
            self.min_validation_loss = validation_loss
            self.counter = 0
        else:
            self.counter += 1
            if self.counter >= self.patience:
                return True
        return False       


# MIT License
# 
# Copyright (c) 2018 Tom Runia
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to conditions.
#
# Author: Tom Runia
# Date Created: 2018-04-16

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np

import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
from mpl_toolkits.axes_grid1 import make_axes_locatable


def plot_scalogram(power, scales, t, normalize_columns=True, cmap=None, ax=None, scale_legend=True):
    """
    Plot the wavelet power spectrum (scalogram).

    :param power: np.ndarray, CWT power spectrum of shape [n_scales,signal_length]
    :param scales: np.ndarray, scale distribution of shape [n_scales]
    :param t: np.ndarray, temporal range of shape [signal_length]
    :param normalize_columns: boolean, whether to normalize spectrum per timestep
    :param cmap: matplotlib cmap, please refer to their documentation
    :param ax: matplotlib axis object, if None creates a new subplot
    :param scale_legend: boolean, whether to include scale legend on the right
    :return: ax, matplotlib axis object that contains the scalogram
    """

    if not cmap: cmap = plt.get_cmap("PuBu_r")
    if ax is None: fig, ax = plt.subplots()
    if normalize_columns: power = power/np.max(power, axis=0)

    T, S = np.meshgrid(t, scales)
    cnt = ax.contourf(T, S, power, 100, cmap=cmap)

    # Fix for saving as PDF (aliasing)
    for c in cnt.collections:
        c.set_edgecolor("face")

    ax.set_yscale('log')
    ax.set_ylabel("Scale (Log Scale)")
    ax.set_xlabel("Time (s)")
    ax.set_title("Wavelet Power Spectrum")

    if scale_legend:
        def format_axes_label(x, pos):
            return "{:.2f}".format(x)
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        plt.colorbar(cnt, cax=cax, ticks=[np.min(power), 0, np.max(power)],
                     format=ticker.FuncFormatter(format_axes_label))

    return ax


import numpy as np
import pandas as pd
import librosa
import cv2 
import pywt 
from ..utils.PyTorchWavelets.wavelets_pytorch.transform import WaveletTransformTorch
from .datahelper import create_map, ana_labels, encoded_labels
# from utils.augmentation import quantile_filter
from sklearn.preprocessing import MinMaxScaler
from copy import deepcopy as dc

def generate_sliding_windows_single(   recording, 
                                        ana_file, 
                                        window_size:int, 
                                        hop_length: int, 
                                        method= 'raw', 
                                        outlier_filter: bool = False, 
                                        scale: bool = True, 
                                        pad_and_slice = True, task = 'train'):
    '''
        ----------
        Arguments
        ----------
            recording: input signal
            ana_file: corresponding analysis file 
            window_size: length of the sliding window
            hop_length: sliding length
            method: signal processing method. Must be one of 'raw', 'fft' or 'spec_img'.
            pad_and_slice: when waveform is shorter than window size, pad the window to 2*window_size and adjust shorter hop length
        --------
        Return
        --------
            (data, label) where
                data: concatenated multidimension arrays of shape (number of windows, shape of the features)
                label: corresponding labels
    '''
    # preprocessing 
    if outlier_filter == True:
        recording = quantile_filter(recording)
    if scale == True:
        # recording = minmax_softscale(recording)
        scaler = MinMaxScaler() # Scale the data to (0,1)
        recording = scaler.fit_transform(recording.reshape(-1,1)).squeeze(1)

    if ana_file is not None:
        ana = dc(ana_file)
        ana['time'] = ana['time'].apply(lambda x: int(x*100))
        ana['time'] = ana['time'].astype(int)

    d = []; l = []
    
    if task == 'train':
        dummy_hop = hop_length     
        for i in range(0,len(ana)-1):

            start = ana['time'][i]
            end = ana['time'][i+1]
            wave_type  = ana['label'][i]
        
            if end-start > window_size: #if longer than window_size, segment into windows 
                pass

            else:#if shorter than window_size, => pad and slice
                if pad_and_slice == True:
                    start, end = (start+end)//2 - window_size, (start+end)//2 + window_size
                    hop_length = 128
                else: 
                    start, end = (start+end)//2 - window_size//2, (start+end)//2 + window_size//2

            n_window = ((end-start)-window_size)//hop_length + 1
            for k in range(n_window):
                if start + k*hop_length + window_size > len(recording):
                    break
                idx = np.arange(start + k*hop_length, start + k*hop_length + window_size)
                window = recording[idx]
                features = extract_features(window, method, window_size)
                d.append(features)
                l.append(wave_type)
            hop_length = dummy_hop

        data = np.stack([w for w in d])
        label = np.array(l)

        return data, label

    elif task == 'test':
        
        n_windows = (len(recording)-window_size)//hop_length + 1
        for i in range(n_windows):
            idx = np.arange(i*hop_length, i*hop_length + window_size)
            window = recording[idx]
            features = extract_features(window, method, window_size)
            d.append(features)

        d = np.array(d)

        if ana_file is not None:
            dense_labels = np.concatenate([[ana['label'].iloc[i]] * (ana['time'].iloc[i+1]-ana['time'].iloc[i])
                            for i in range(len(ana)-1)]) 
            map = create_map(ana_labels, encoded_labels)
            dense_labels = pd.Series(dense_labels).map(map).to_numpy()
            return d, dense_labels
        else:
            return d

# def extract_features(window, method, window_size):
#     if method == 'fft':
#         features = (np.abs(librosa.stft(window,n_fft=window_size,center=False))/np.sqrt(window_size)).ravel()
#     elif method == 'spec_img':
#         features = librosa.amplitude_to_db(np.abs(librosa.stft(window, n_fft=128, center = False, hop_length=14))/np.sqrt(window_size),ref=np.max)
#     elif method == 'raw':
#         features = window
#     else:
#         raise RuntimeError ("Param 'method' should be one of 'fft',  'spec_img' or 'raw'.")
#     return features

def extract_features(windows, method, window_size):
    if method == 'raw':
        features = windows
    elif method == 'fft':
        features = (np.abs(librosa.stft(windows,n_fft=window_size,center=False))/np.sqrt(window_size)).squeeze(-1)
    elif method == 'spectrogram':
        features = librosa.amplitude_to_db(np.abs(librosa.stft(windows, n_fft=128, center = False, hop_length=14))/np.sqrt(window_size),ref=np.max)
    elif method == 'gaf':
        gaf = GAF(image_size=64, method = 'summation', overlapping=True)
        features = gaf.transform(windows)
    elif method == 'wavelet':
        a3, d3, d2, d1 = pywt.wavedec(windows,'sym4',level = 3)
        n = windows.shape[0]
        d1_shape = d1.shape[1]
        a3 = interpolate(a3, d1_shape)
        d3 = interpolate(d3, d1_shape)
        d2 = interpolate(d2, d1_shape)

        features = np.stack([np.stack([a3[i],d3[i],d2[i],d1[i]]) for i in range(a3.shape[0])])

    elif method == 'scalogram':
        torch_wt = WaveletTransformTorch(0.01, 9/64, cuda = True)
        features = np.abs(torch_wt.cwt(windows))
        resized = []
        for i in range(features.shape[0]):
            im = features[i]
            resized_scalogram = cv2.resize(im, (im.shape[0], im.shape[0]), interpolation = cv2.INTER_LINEAR)
            resized.append(resized_scalogram)
        features = np.stack(resized)
    else:
        raise RuntimeError ("Param 'method' should be one of 'fft',  'spectrogram' or 'raw'.")
    return features

def interpolate(array, n_interp):
    ndim = len(array.shape)
    if ndim == 1:
        return np.interp(np.arange(0, n_interp), np.arange(0, len(array)), array)
    else:
        n = array.shape[0]
        interp_array = []
        for i in range(n):
            tmp = np.interp(np.arange(0, n_interp), np.arange(0, len(array[i,:])), array[i,:])
            interp_array.append(tmp)
        return np.array(interp_array)


from sklearn.model_selection import train_test_split, cross_validate
from sklearn.metrics import accuracy_score

import numpy as np
import pandas as pd
import pywt
from scipy.stats import skew

from copy import deepcopy as dc
import matplotlib.pyplot as plt

import time
import datetime
from easydict import EasyDict
import os
from tqdm import tqdm 

from .Dataset import EPGDataset
from ..dataset_utils.datagenerator import generate_sliding_windows_single
from ..dataset_utils.datahelper import read_signal
from ..utils import metrics, visualization
from ..utils.stats import calculate_statistics

# ======================= MODEL ===========================
from sklearn.ensemble import RandomForestClassifier
from xgboost import XGBClassifier
from sklearn.svm import SVC
from sklearn.linear_model import LogisticRegression

def get_MLmodel(name):
    if name == 'RF':
        return RandomForestClassifier(class_weight='balanced', n_jobs=-1, random_state=28)
    elif name == 'XGB':
        return XGBClassifier(n_jobs = -1, n_estimators = 300, max_depth = 6, eta = 0.3)
    elif name == 'SVC':
        return SVC(class_weight = 'balanced', kernel = 'linear')
    elif name == 'LogReg':
        return LogisticRegression(n_jobs = -1, penalty= 'l1', solver = 'saga', warm_start=True)
    else:
        raise ValueError("Unsupported model. Must be one of ['XGB', 'RF', 'SVC', 'LogReg']")

# For Grid Search on XGB
XGB_params_grid = {'eta': [0.01, 0.1, 0.2, 0.3], 'n_estimators': [50,100, 200, 300],'max_depth': [3,4,5,6]}

class EPGSegmentML():
    def __init__(self, config, inference = False):

        # Dataset
        self.data_path = config.data_path 
        self.dataset_name = config.dataset_name
        self.dataset = EPGDataset(self.data_path, self.dataset_name)

        # Model/ optimizers
        self.config = config
        self.device = config.device        
        self.model = get_MLmodel(config.arch)

        # Configs for inputs
        self.window_size = config.window_size
        self.hop_length = config.hop_length
        self.scope = config.scope
        self.method = config.method 
        self.scale = config.scale 

        # Environment
        self.random_state = 28
        self._is_model_trained = False
        self._is_pretrained = False
        self.processed_data_path = f'{self.data_path}/dataML'
        # Result
        self.classification_result_ = EasyDict({})

    def reset(self):
        self.model = get_MLmodel(config.arch)
        self.classification_result_ = EasyDict({})

    def generate_train_data(self, test_size = 0.2):
        print(f'Processed dataset is saved to {self.processed_data_path}')
        if os.path.exists(f'{self.data_path}/dataML/Label_{self.dataset_name}_train.csv'):
            print('Warning. Training data existed.')
            inp = input('Continue? (Y/N)')
            if inp == 'Y':
                pass
            elif inp == 'N':
                return 
        self.dataset.generate_sliding_windows(  window_size     = self.window_size, 
                                                hop_length      = self.hop_length, 
                                                scale           = self.scale, 
                                                method          = 'raw', 
                                                # outlier_filter = False, 
                                                # pad_and_slice = True, 
                                                verbose         = True)
        data, labels = self.dataset.windows, self.dataset.labels

        # Compute features
        t0 = time.perf_counter() # Time counter
        
        X_train, X_test, y_train, y_test = train_test_split(data, labels, test_size = 0.2, random_state = 28, stratify = labels)

        print('Computing training features matrices ...')
        X_train = calculate_features(X_train, method = self.method)

        print('Computing testing features matrices ...')
        X_test = calculate_features(X_test, method = self.method)

        t1 = time.perf_counter() # Time counter

        print(f'Dataset {self.dataset_name}. Elapsed computation time: {t1-t0}')

        # Save as .csv
        print('Saving datasets to .csv ...')
        os.makedirs(f'{self.data_path}/dataML', exist_ok = True)
        pd.DataFrame(y_train).to_csv(f'{self.data_path}/dataML/Label_{self.dataset_name}_train.csv',header = None, index= None)
        pd.DataFrame(y_test).to_csv(f'{self.data_path}/dataML/Label_{self.dataset_name}_test.csv',header = None, index= None)
        pd.DataFrame(X_train).to_csv(f'{self.data_path}/dataML/Data_{self.dataset_name}_train.csv',header = None, index= None)
        pd.DataFrame(X_test).to_csv(f'{self.data_path}/dataML/Data_{self.dataset_name}_test.csv',header = None, index= None)
        
        # f = open(f'{self.data_path}/log/features_computation_time.txt', 'w')
        # f.writelines([f'Dataset {self.dataset_name}. Elapsed features computation time: {t1-t0}\n'])

    def load_train_data(self, data_path = None, X_train = None, X_test = None, y_train = None, y_test = None):
        if (X_train is None) and (X_test is None) and (y_train is None) and (y_test is None):
            if data_path is None:
                data_path = self.processed_data_path
            self.X_train, self.X_test, self.y_train, self.y_test = read_dataset_csv(self.dataset_name, data_path)   
            print('Data loaded from', data_path) 
        else:
            if (X_train is not None) and (X_test is not None) and (y_train is not None) and (y_test is not None):
                self.X_train, self.X_test, self.y_train, self.y_test = X_train, X_test, y_train, y_test
            else:
                raise RuntimeError('Must input all of X_train, X_test, y_train, y_test.')

    def fit(self):
        #TODO: Fix cases with testing labels miss some classes. Perhaps creating a new label map?
        print('Training...')
        t0 = time.perf_counter()
        self.model.fit(self.X_train, self.y_train)
        t1 = time.perf_counter()
        training_time = round(t1 - t0, 3)
        self.classification_result_['training_time'] = training_time
        print(f'Finished training. Elapsed time: {training_time} (s)')

        self._is_model_trained = True 

    def predict(self, X_test, y_test):

        if self._is_model_trained == False:
            raise Warning('Model is not trained.')

        # Prediction
        print('Predicting ...')
        self.pred_proba = self.model.predict_proba(X_test)
        pred_windows = np.argmax(self.pred_proba, axis = -1)
        # Scoring
        results = metrics.scoring(y_test, pred_windows)
        scores = results['scores']
        cf = results['confusion_matrix']
        self.classification_result_['class_acc'] = [cf[i,i] for i in range(cf.shape[0])]
        self.classification_result_['test_scores'] = scores
        self.classification_result_['test_confusion_matrix'] = cf
        
        print(f"Accuracy: {scores['accuracy']}, f1: {scores['f1']}")
        print('Finished testing.')

    def cross_validate(self, cv = 10, verbose = 0, keep_best_model = True):
        scores = ['accuracy','f1_macro','precision_macro','recall_macro']
        self.cv_results = cross_validate(self.model, self.X_train, self.y_train, scoring = scores, cv=cv, n_jobs=-1,verbose = verbose, return_estimator= True)
        cv_scores = list(self.cv_results.keys())
        cv_scores.remove('estimator')
        summary = {}
        for key in cv_scores:
            summary[key] = [np.round(np.mean(self.cv_results[key]),2), np.round(np.std(self.cv_results[key]),2)]
        best_estimator_index = np.argmax(self.cv_results['test_accuracy'])
        best_estimator = self.cv_results['estimator'][best_estimator_index]
        if keep_best_model == True:
            self.model = best_estimator 
        self.cv_summary = summary

    def write_cv_log(self):

        date = str(datetime.datetime.now())[:-7]
        os.makedirs(f'log/{self.config.arch}', exist_ok = True)
        with open(f'log/{self.config.arch}/cv_results.txt','a') as f:
            f.writelines([f'Date: {date} | Model: {self.config.arch} | Dataset: {self.config.dataset_name}\n'])
            cv_summary_txt = [f'{key}: {self.cv_summary[key][0]} +- {self.cv_summary[key][1]}\n' for key in self.cv_summary.keys()]
            f.writelines(cv_summary_txt)

    def write_test_result(self):
        date = str(datetime.datetime.now())[:-7]
        os.makedirs(f'log/{self.config.arch}', exist_ok = True)        
        with open(f'log/{self.config.arch}/test_results.txt','a') as f:
            f.writelines([f'Date: {date} | Model: {self.config.arch} | Dataset: {self.config.dataset_name}\n'])
            f.writelines([f"{s}: {self.classification_result_['test_scores'][s]} \n" for s in self.classification_result_['test_scores'].keys()])  
            f.writelines(f'Class accuracy: {self.classification_result_["class_acc"]}\n')  

    def segment(self, recording_name, is_FS = False, return_score = True):

        # Prepare data
        print('Generating segmentation ...')
        self.recording_name = recording_name
        self.recording, self.ana = read_signal(recording_name)
        test_hop_length = self.hop_length // self.scope
        data = generate_sliding_windows_single(self.recording, self.ana, 
                                            window_size = self.window_size, 
                                            hop_length = test_hop_length, 
                                            method = 'raw', 
                                            scale = self.scale, 
                                            task = 'test')
        if self.ana is not None:
            self.input = calculate_features(data[0], method = self.config.method)
            self.true_segmentation = data[1]
        else:
            self.input = calculate_features(input, method = self.config.method)
        selected_features = [5,6,9,10,11,12,17,18,22,23,24,25,30,31,35,36,37,38,43,44,48,49,50,51]
        if is_FS == True:
            self.input = self.input[:, selected_features]
        # Predict
        

        self.pred_segm_proba = self.model.predict_proba(self.input)
        pred_windows = np.argmax(self.pred_segm_proba, axis = -1)

        # Generate segmentation
        pred_segmentation = []
        for i in range(len(pred_windows)):
            pred_segmentation.extend([pred_windows[i]]*test_hop_length)
        pred_segmentation = np.array(pred_segmentation)
        pred_segmentation = extend(pred_segmentation, self.true_segmentation)

        # Calculate overlapping rate
        accuracy = accuracy_score(self.true_segmentation, pred_segmentation) 
        self.overlap_rate = accuracy
        print(f"Overlapping rate: {accuracy}")

        # map to ground_truth labels  
        self.pred_segmentation = pd.Series(pred_segmentation).map({0: 1, 1: 2, 2: 4, 3: 5, 4: 6, 5: 7, 6: 8}).to_numpy() 
        self.pred_ana = to_ana(self.pred_segmentation) 
        self.pred_ana =self.pred_ana[['label', 'time']]
        if return_score == True:
            return self.pred_ana, self.overlap_rate
        else:
            return self.pred_ana

    def save_analysis(self, name: str = ''):

        os.makedirs('./prediction/ANA', exist_ok = True)
        dir = os.listdir('./prediction/ANA')
        index = len(dir) + 1
        if name == '':
            self.pred_ana.to_csv(f'./prediction/ANA/{self.recording_name}.ANA',sep = '\t',header = None,index=None)
        else:
            self.pred_ana.to_csv(f'./prediction/ANA/{name}.ANA',sep = '\t',header = None,index=None)

    def plot_pred_proba(self, r: tuple = None, ax = None):
        visualization.plot_pred_proba(self.pred_segm_proba, self.config.hop_length, self.config.scope, r, ax)

    def plot_segmentation(self, which = 'pred_vs_gt', hour = None, range = None, savefig = False): 
        visualization.plot_gt_vs_pred_segmentation(self.recording, self.ana, self.pred_ana, hour = hour, range = range, which = which, savefig = savefig, name = self.recording_name)
        
    def interactive_plot(self, which = 'prediction', smoothen = False):
        if which == 'prediction':
            visualization.interactive_visualization(self.wave_array, self.predicted_analysis, smoothen, title = which)
        elif which == 'ground_truth':
           visualization.interactive_visualization(self.wave_array, self.ana, smoothen, title = which)
        else:
            raise RuntimeError("Must input either 'prediction' or 'ground_truth' ")

### Some utilities function      ##########################
def calculate_features(df, method):
    d = []
    for i in tqdm(range(len(df))):
        feats_vector = []
        if method == 'wavelet':
            coef = pywt.wavedec(df[i,:],'sym4',level = 3)
            for i in range(len(coef)):
                feats = calculate_statistics(coef[i])
                feats_vector += feats
        elif method == 'fft':
            n_fft = df[i,:].shape[0]
            coef = (np.abs(librosa.stft(df[i, :],n_fft=n_fft,center=False))/np.sqrt(n_fft)).ravel()
            feats_vector += calculate_statistics(coef)
        elif method == 'raw':
            feats_vector += calculate_statistics(df[i,:])
        else: 
            raise ValueError(f"Undefined method {str(method)}. Must be 'wavelet','fft' or 'raw'.")
        d.append(feats_vector)
    d = np.stack([f for f in d])
    return d

def to_ana(segmentation):
    # create *.ANA file
    ana_time = [0]
    ana_label = []
    tmp_label = segmentation[0]
    for i in range(1, len(segmentation)):
        if segmentation[i] != segmentation[i-1]: 
            ana_label.append(tmp_label)
            ana_time.append(i/100)
            tmp_label = segmentation[i]
    ana_label += [segmentation[i], 99]
    ana_time += [28800]
    pred_ana = pd.DataFrame({'time': ana_time, 'label': ana_label})
    
    return pred_ana

def repeat(array, coef):
    shape = array.shape 
    if len(shape) == 3:
        if shape[1] == 1:
            tmp = array.reshape(([shape[0],shape[2]]))
    n = len(array)
    repeated = []
    for i in range(n):
        repeated.extend([array[i]] * coef)
    return np.array(repeated)

def extend(arr1, target):
    '''
        Repeat the last axis of arr1 until 
        as long as target array (or reach target length)
    '''
    if isinstance(target, np.ndarray):
        ext_len = len(target) - len(arr1)
    else: 
        ext_len = target - len(arr1)
    if ext_len < 0:
        raise RuntimeError('Cannot repeat. Target length is less than input length')
    if len(arr1.shape) == 1: 
        extension = np.repeat(arr1[-1], ext_len)
    else:
        extension = np.tile(arr1[-1,:], (ext_len,1))
    return np.concatenate([arr1, extension])

def read_dataset_csv(dataset_name, data_path):
    columns = []
    for i in range(4):
        columns += [f'n5_{i}', f'n25_{i}', f'n75_{i}', f'n95_{i}', f'median_{i}', 
                    f'mean_{i}', f'std_{i}', f'var_{i}', f'rms_{i}', f'sk_{i}', f'zcr_{i}', f'en_{i}', f'perm_en_{i}']
    X_train = pd.read_csv(f'{data_path}/Data_{dataset_name}_train.csv',header = None)
    X_test = pd.read_csv(f'{data_path}/Data_{dataset_name}_test.csv',header = None)
    y_train =  pd.read_csv(f'{data_path}/Label_{dataset_name}_train.csv',header = None)
    y_test =  pd.read_csv(f'{data_path}/Label_{dataset_name}_test.csv',header = None)
    
    X_train.columns = columns
    X_test.columns = columns
    return X_train, X_test, y_train, y_test


import argparse


def get_args():
    argparser = argparse.ArgumentParser(description=__doc__)
    argparser.add_argument(
        '-c', '--config',
        metavar='C',
        default='None',
        help='The Configuration file',
        required=True)
    args = argparser.parse_args()
    return args


# MIT License
# 
# Copyright (c) 2018 Tom Runia
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to conditions.
#
# Author: Tom Runia
# Date Created: 2018-04-16

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import matplotlib.pyplot as plt

from wavelets_pytorch.wavelets import Morlet, Ricker, DOG
from wavelets_pytorch.transform import WaveletTransform, WaveletTransformTorch
from examples.plot import plot_scalogram

"""
Example script to plot SciPy and PyTorch implementation outputs side-to-side.
"""

fps = 20
dt  = 1.0/fps
dj  = 0.125
unbias = False
batch_size = 32
wavelet = Morlet(w0=6)

t_min = 0
t_max = 10
t = np.linspace(t_min, t_max, (t_max-t_min)*fps)

######################################
# Generating batch of random sinusoidals

random_frequencies = np.random.uniform(0.5, 4.0, size=batch_size)
batch  = np.asarray([np.sin(2*np.pi*f*t) for f in random_frequencies])
batch += np.random.normal(0, 0.2, batch.shape)  # Gaussian noise

######################################
# Performing wavelet transform

wa = WaveletTransform(dt, dj, wavelet, unbias=unbias)
wa_torch = WaveletTransformTorch(dt, dj, wavelet, unbias=unbias, cuda=True)

power = wa.power(batch)
power_torch = wa_torch.power(batch)

######################################
# Plotting

fig, ax = plt.subplots(1, 3, figsize=(12,3))
ax = ax.flatten()
ax[0].plot(t, batch[0])
ax[0].set_title(r'$f(t) = \sin(2\pi \cdot f t) + \mathcal{N}(\mu,\,\sigma^{2})$')
ax[0].set_xlabel('Time (s)')

# Plot scalogram for SciPy implementation
plot_scalogram(power[0], wa.fourier_periods, t, ax=ax[1], scale_legend=False)
ax[1].axhline(1.0 / random_frequencies[0], lw=1, color='k')
ax[1].set_title('Scalogram (SciPy)'.format(1.0/random_frequencies[0]))

# Plot scalogram for PyTorch implementation
plot_scalogram(power_torch[0], wa_torch.fourier_periods, t, ax=ax[2])
ax[2].axhline(1.0 / random_frequencies[0], lw=1, color='k')
ax[2].set_title('Scalogram (Torch)'.format(1.0/random_frequencies[0]))
ax[2].set_ylabel('')
ax[2].set_yticks([])

plt.tight_layout()
plt.show()


import numpy as np
import pandas as pd 
from sklearn.preprocessing import MinMaxScaler
import warnings

import os
from tqdm import tqdm
import time

import matplotlib.pyplot as plt 
from ipywidgets import widgets, interact 

from ..dataset_utils.datahelper import format_data, read_signal, get_index
from ..dataset_utils.datagenerator import generate_sliding_windows_single
from ..utils.visualization import visualize_signal, interactive_visualization


class EPGDataset:
    def __init__(
                self, 
                data_path = '../data', 
                dataset_name = 'SA',
                ):

        self.data_path = data_path
        self.dataset_name = dataset_name
        self._format_recNames()

        self.recNames = os.listdir(f'{self.data_path}/{self.dataset_name}')
        self.recNames = sorted(set([x[:-4] for x in self.recNames]))
        t = time.perf_counter()
        self.recordings = []
        print('Loading data ...')
        for id, recording_name in enumerate(tqdm(self.recNames)):
            recording, ana = read_signal(recording_name, data_path=self.data_path)
            self.recordings.append({'id': id,
                                    'name': recording_name,
                                    'recording': recording,
                                    'ana':ana    
                                    })
        print(f'Done! Elapsed: {time.perf_counter() - t} s')

        self.guidelines = [
                            'The recording is always started with NP',
                            'Np is always followed by C or last until the end',
                            'E2 is always preceded by E1',
                            'pd waveform is always preceded by C and followed by C or Np',
                            'The subphase sequence after pd (pdII-1) should be II-2, II-3',
                            'F is always preceded by C',
                            'F is always followed by Np or C',
                            'G is always preceded by C',
                            'G is always followed by Np or C',
                            'E1 is always followed by E1e, E2, C, or Np',
                            'The end of each recording must be marked (T, code 99)',
                            ]        

    def __len__(self):
        return len(self.recordings)

    def __getitem__(self, idx):
        return self.recordings[idx]

    def loadRec(self, recName):
        if isinstance(recName, str):
            idx = self.recNames.index(recName)
            return self.__getitem__(idx)
        elif isinstance(recName, list):
            recs = []
            for name in recName:
                idx = self.recNames.index(name)
                recs.append(self.__getitem__(idx))
            return recs

    def _format_recNames(self):
        recNames = os.listdir(f'{self.data_path}/{self.dataset_name}')
        prefix = f'{self.dataset_name}_'
        formated = 0
        for name in recNames:
            if not name.startswith(prefix):
                formated = 1
                newname = f'{prefix}{name}'
                os.rename(f'{self.data_path}/{self.dataset_name}/{name}',f'{self.data_path}/{self.dataset_name}/{newname}')
        anaNames = os.listdir(f'{self.data_path}/{self.dataset_name}_ANA')
        for name in anaNames:
            if not name.startswith(prefix):
                newname = f'{prefix}{name}'
                os.rename(f'{self.data_path}/{self.dataset_name}_ANA/{name}',f'{self.data_path}/{self.dataset_name}_ANA/{newname}')
        if formated == 1:
            print(f'Filenames formated with prefix {prefix}.')
            
    def plot(self, idx, mode = 'static', hour = None, range = None, width = None, height = None, smoothen = False):
        if isinstance(idx, str):
            recData = self.loadRec(idx)
        elif isinstance(idx, int):
            recData = self.recordings[idx]
        recording, ana = recData['recording'], recData['ana']
        plt.figure(figsize = (18,3))
        if mode == 'static':
            if width is not None or height is not None:
                warnings.warn("'width' and 'height' arguments are only applicable for interactive plot.")
            if smoothen == True:
                warnings.warn("'smoothen' arguments are only applicable for interactive plot.")
            visualize_signal(recording, ana, title = recData['name'], hour = hour, range = range)
        elif mode == 'interactive':
            interactive_visualization(recording, ana, smoothen= smoothen, hour = hour, range = range, 
                                        width = width, height = height, title = recData['name'])

    def generate_sliding_windows(   self,
                                    window_size = 1024, 
                                    hop_length = 1024, 
                                    method= 'raw', 
                                    outlier_filter: bool = False, 
                                    scale: bool = True, 
                                    pad_and_slice = True, 
                                    verbose = True):
        '''
            ----------
            Arguments
            ----------
                config: configuration files containing necessary info
                verbose: if True, print descriptions
            --------
            Return
            --------
                d: dictionaries of training/testing data with keys {'data', 'label'}
        '''
        print('Generating sliding windows ...')
        count = 0
        d = []; l = []  
        
        for rec in tqdm(self.recordings):
            recRecording = rec['recording']
            recAna = rec['ana']
            features, labels = generate_sliding_windows_single(recRecording, recAna, window_size, hop_length, 
                                                                method, outlier_filter, scale, True, 'train')
            d.append(features); l.append(labels)
            count+=1

        d = np.concatenate([f for f in d])
        l = np.concatenate([lab for lab in l])
        
        d = format_data(d, l)

        if verbose == True:
            print(f'Total: {count} recordings')
            print(f'Signal processing method: {method} | Scale: {str(scale)}')
            cl, c = np.unique(l, return_counts=True)
            print('Class distribution (label:ratio): '+ ', '.join(f'{cl[i]}: {round(c[i]/len(l),2)}' for i in range(len(cl))))
            print(f'Labels map (from:to): {{1: 0, 2: 1, 4: 2, 5: 3, 6: 4, 7: 5, 8: 6}}')

        self.windows, self.labels = d['data'], d['label']
        self.waveforms, self.distributions = np.unique(self.labels, return_counts= True)
        n = len(self.waveforms)
        self.distributions = [round(self.distributions[i]/len(self.labels),2) for i in range(n)]
        self.label_to_id = {1: 0, 2: 1, 4: 2, 5: 3, 6: 4, 7: 5, 8: 6}
        self.name_to_label = {'NP':1, 'C':2, 'E1':4, 'E2':5, 'F':6, 'G':7, 'pd':8}
        return self.windows, self.labels 

    def check_guidelines(self, ana):

        # 1. The recording is always started with NP
        check = [False]*len(self.guidelines)
        check_log = ['']*len(self.guidelines)
        if ana.iloc[0,0] == 1:
            check[0] = True
            check_log[0] = '1. Checked'
        else:
            check_log[0] = '1. Recording does not start with NP'
        # 2. Np is always followed by C or last until the end
        NP_idx = ana.index[ana['label'] == 1]
        next_idx = NP_idx + 1
        next_label = ana.loc[next_idx]['label']
        next_notC_idx = next_label.index[~(next_label.isin([2,12,99]))]
        if len(next_notC_idx) == 0:
            check[1] = True 
            check_log[1] = '2. Checked'
        else:
            check_log[1] = f'2. At line {str(list(next_notC_idx-1))[1:-1]}, NP is not followed by C.'
        
        # 3. E2 is always preceded by E1
        E2_idx = ana.index[ana['label'] == 5]
        prev_idx = E2_idx - 1
        prev_label = ana.loc[prev_idx]['label']
        prev_notE1_idx = prev_label.index[prev_label != 4]
        if len(prev_notE1_idx) == 0:
            check[2] = True 
            check_log[2] = '3. Checked'
        else:
            check_log[2] = f'3. At line {str(list(prev_notE1_idx+1))[1:-1]}, E2 is not preceeded by E1.'
        
        # 4. pd waveform is always preceded by C and followed by C or Np
        pd_idx = ana.index[ana['label'] == 8]
        not_prev_by_C = []
        not_follw_by_C_or_NP = []
        for idx in pd_idx:
            if ana.loc[idx-1,'label'] != 2:
                not_prev_by_C.append(idx)
            else:
                if ana.loc[idx+1, 'label'] != 2 and ana.loc[idx+1, 'label'] != 1:
                    not_follw_by_C_or_NP.append(idx)
        invalid_idx = not_prev_by_C + not_follw_by_C_or_NP
        if len(invalid_idx) != 0:
            check_log[3] = f'4. At line {str(list(not_prev_by_C))[1:-1]}, pd is not preceeded by C.', +\
                            f'At line {str(list(not_follw_by_C_or_NP))[1:-1]}, pd is not followed by C or NP.'
        else:
            check[3] = True
            check_log[3] = '4. Checked.'
        # 5. The subphase sequence after pd (pdII-1) should be II-2, II-3
        allWaveformIds = ana['label'].unique()
        if 9 not in allWaveformIds or 10 not in allWaveformIds:
            check[4] = True 
            check_log[4] = '5. Checked. pd-II-2, pd-II-3 is not presented.'
        # 6. F is always preceded by C and 7. F is always followed by Np or C
        F_idx = ana.index[ana['label'] == 6]
        if len(F_idx) > 0:
            not_prev_by_C = []
            not_follw_by_C_or_NP = []
            for idx in F_idx:
                if ana.loc[idx-1,'label'] != 2:
                    not_prev_by_C.append(idx)
                else:
                    if ana.loc[idx+1, 'label'] != 2 and ana.loc[idx+1, 'label'] != 1:
                        not_follw_by_C_or_NP.append(idx)
            if len(not_prev_by_C) != 0: # 6. F is always preceded by C
                check_log[5] = f'6. At line {str(list(not_prev_by_C))[1:-1]}, F is not preceeded by C.'
            else:
                check[5] = True
                check_log[5] = '6. Checked.'
            if len(not_follw_by_C_or_NP) != 0: #7. F is always followed by Np or C
                check_log[6] = f'7. At line {str(list(not_follw_by_C_or_NP))[1:-1]}, F is not followed by C or NP.'
            else:
                check[6] = True
                check_log[6] = '7. Checked.'
        else:
            check[5] = True
            check_log[5] = '6. Checked. F is not presented.'
            check[6] = True    
            check_log[6] = '7. Checked. F is not presented.'
        # 8. F is always preceded by C and 9. F is always followed by Np or C
        G_idx = ana.index[ana['label'] == 6]
        if len(G_idx) > 0:
            not_prev_by_C = []
            not_follw_by_C_or_NP = []
            for idx in G_idx:
                if ana.loc[idx-1,'label'] != 2:
                    not_prev_by_C.append(idx)
                else:
                    if ana.loc[idx+1, 'label'] != 2 and ana.loc[idx+1, 'label'] != 1:
                        not_follw_by_C_or_NP.append(idx)
            if len(not_prev_by_C) != 0: # 8. G is always preceded by C
                check_log[7] = f'8. At line {str(list(not_prev_by_C))[1:-1]}, G is not preceeded by C.'
            else:
                check[7] = True
                check_log[7] = '8. Checked.'
            if len(not_follw_by_C_or_NP) != 0: #9. G is always followed by Np or C
                check_log[8] = f'9. At line {str(list(not_follw_by_C_or_NP))[1:-1]}, G is not followed by C or NP.'
            else:
                check[8] = True
                check_log[8] = '9. Checked.'
        else:
            check[7] = True
            check_log[7] = '8. Checked. G is not presented.'
            check[8] = True  
            check_log[8] = '9. Checked. G is not presented.'
        # 10. E1 is always followed by E1e, E2, C, or Np
        E1_idx = ana.index[ana['label'] == 4]
        next_idx = E1_idx - 1
        next_label = ana.loc[next_idx]['label']
        next_notE1eE2CNP_idx = next_label.index[~(next_label.isin([3, 5, 2, 1]))]
        if len(next_notE1eE2CNP_idx) == 0:
            check[9] = True 
            check_log[9] = '10. Checked.'
        else:
            check_log[9] = f'10. At line {str(list(next_notE1eE2CNP_idx+1))[1:-1]}, E1 is not followed by E1e, E2, C, or Np.'
                
        # 11. The end of each recording must be marked (T, code 99)
        last_row = ana.iloc[-1,:]
        if last_row['label'] == 12 or last_row['label'] == 99:
            check[10] = True
            check_log[10] = '11. Checked.'
        else:
            check_log[10] = '11. The recording does not end with label 99 or 12.'
        
        check_text = [f'{i+1}. {self.guidelines[i]}: {check[i]}' for i in range(len(self.guidelines))]

        return check_text, check_log
        
    def getRecordingParams(self, idx = None, ana = None, progress = 'overall', export_xlsx = True):
        
        if idx == None:
            idx = [i for i in range(len(self.recordings))]
        else:
            if isinstance(idx, int):
                idx = [idx]
        self.params = []
        self.error_log = {}
        self.guideline_check_log = []
        for recIndex in idx:
            recData = self.recordings[recIndex]
            recName = recData['name']
            recId = recData['id']
            recRecording = recData['recording']
            if ana is None:
                recAna = recData['ana']
                # print(recAna)
            else:
                recAna = ana
            recTime = recAna.iloc[-1,1]//3600
            waveformIndices = get_index(recAna)
            # allWaveformIds = set(recAna.loc[:,'label'])
            listWaveforms = list(waveformIndices.keys())
            error_log = []
            params = {}
            params['recTime'] = recTime
            params['waveforms'] = ', '.join(listWaveforms)
            
            ############ NP (non-probing) #############

            if 'NP' in listWaveforms:
                params['s_NP']  = 0 
                params['n_NP']    = 0 
                params['a_NP']   = 0   
                params['m_NP'] = 0   
                params['mx_NP']    = 0     
                # for i in range(1, int(recTime) + 1):
                #     params[f's_NP_h{i}'] = 0     
                error_log.append('No NP waveform found.')                  
            else:
                NP_durations        = [x[1] - x[0] for x in waveformIndices['NP']]
                params['s_NP']  = float(np.sum(NP_durations).astype(np.float32))
                params['n_NP']    = len(waveformIndices['NP'])
                params['a_NP']   = float(np.mean(NP_durations).astype(np.float32))
                params['m_NP'] = float(np.median(NP_durations).astype(np.float32))  
                params['mx_NP']    = float(np.max(NP_durations).astype(np.float32))
                if len(NP_durations) < 2:
                    params['t_2nd_NP'] = 0 
                else:
                    params['t_2nd_NP'] = NP_durations[1]
                
            
            # for i in range(1, int(recTime) + 1):
            #     NP_hour_i = [x for x in waveformIndices['NP'] if x[0] > i*3600. and x[1] <= (i+1)*3600.] 
            #     params[f's_NP_h{i}'] = float(np.sum([x[1] - x[0] for x in NP_hour_i]).astype(np.float32))
            # # params[f'n_NP_h{i}'] = len(NP_hour_i)            
            #     
            ############ C, pd, probes ############# (Assuming that a recording always have C -> always have probes)
            if 'C' not in listWaveforms:
                error_log.append('No C waveform found.')
            probe_locations = get_stage(recAna, stage = 'probe')
            probes_durations        = [x[1] - x[0] for x in probe_locations]
            params['s_Pr']  =   float(np.sum(probes_durations).astype(np.float32))
            params['n_Pr']    =   len(probes_durations)
            params['a_Pr']   =   float(np.mean(probes_durations).astype(np.float32))
            params['m_Pr'] =   float(np.median(probes_durations).astype(np.float32))

            # Total duration of C, Number of C, Mean duration of C 
            C_durations         = [x[1] - x[0] for x in waveformIndices['C']]
            params['s_C']   = float(np.sum(C_durations).astype(np.float32))
            params['n_C']     = len(C_durations)
            params['a_C']    = float(np.mean(C_durations).astype(np.float32))
            params['m_C']  = float(np.median(C_durations).astype(np.float32))
            params['%C_in_Pr']  = params['s_C'] / params['s_Pr']

            # Time from start of the experiment to 1st probe
            params['t>1stPr'] = waveformIndices['C'][0][0]

            # Duration of 1st probe, Duration of 2nd probe
            params['t_1stPr'] = probe_locations[0][1] - probe_locations[0][0]
            if len(probe_locations) <= 1:
                params['t_2ndPr'] = 0    
                error_log.append(f'Duration of 2nd probes = 0 as number of probes = {len(probe_locations)}')
            else:    
                params['t_2ndPr'] = probe_locations[1][1] - probe_locations[1][0]

            # Number of short probes (<3 min)
            shortprobe_locations = [x for x in probe_locations if x[1] - x[0] < 180 and x[1] - x[0] > 0]
            params['n_shortPr'] = len(shortprobe_locations)
            # Number of very short probes (<1 min)
            veryshortprobe_locations = [x for x in probe_locations if x[1] - x[0] < 60 and x[1] - x[0] > 0]
            params['n_veryshortPr'] = len(veryshortprobe_locations)     
            # Number of probes, Number of probes during the 1st, 2nd, ..., 8th h, Total probing time
            
            # for i in range(1, int(recTime) + 1):
            #     probes_hour_i              = [x for x in probe_locations if x[0] > i*3600. and x[1] <= (i+1)*3600.] 
            #     params[f'n_probes_h{i}'] = len(probes_hour_i)


            # Number of pd, Mean duration of pd
            if 'pd' not in listWaveforms:
                params['s_pd']  = 0 
                params['a_pd']   = 0 
                params['n_pd']    = 0 
                params['m_pd'] = 0 
                params['t>1stpd']  = 0 

                # params['t_last_pd_>end'] = 0  # Time from the end of the last pd to the end of the EPG record (Z)

            else:
                pd_durations        = [x[1] - x[0] for x in waveformIndices['pd']]
                params['s_pd']  = float(np.sum(pd_durations).astype(np.float32))
                params['n_pd']    = len(pd_durations)
                params['a_pd']   = float(np.mean(pd_durations).astype(np.float32))
                params['m_pd'] = float(np.median(pd_durations).astype(np.float32))
                params['t>1stpd']  = waveformIndices['pd'][0][0]
            #     params['t_last_pd_>end'] = len(recRecording)//100 - waveformIndices['pd'][-1][0] 
            # Number of probes to the 1st E1
            # Time from the end of the last pd to the end of the probe

            # Time from 1st probe to 1st pd
            if 'pd' not in listWaveforms:
                params['t_start1stPr>1stpd'] = 0 
            else:
                params['t_start1stPr>1stpd'] = waveformIndices['pd'][0][0] - probe_locations[0][0]
                        

            # # Time from the start of E2 to the end of the EPG record (Z)
            # if 'E2' not in listWaveforms:
            #     params['t_startE2>end'] = 0 
            # else:
            #     params['t_startE2>end'] = len(recRecording)//100 - waveformIndices['E2'][0][0]

            # # Time from start of the experiment to 1st E2
            # if 'E2' not in listWaveforms:
            #     params['t>1stE2'] = 0 
            # else:
            #     params['t>1stE2'] = waveformIndices['E2'][0][0]

            ############ F #############
            # Total duration of F, Total duration of F during the 1st, 2nd, ..., 8th h, 
            # Number of F, Number of F during the 1st, 2nd, ..., 8th h, Mean duration of F
            if 'F' not in listWaveforms:
                params['s_F']   = 0 
                params['a_F']    = 0 
                params['n_F']     = 0 
                params['m_F']  = 0 
                params['%F_in_Pr']  = 0 

            else:
                F_durations         = [x[1] - x[0] for x in waveformIndices['F']]
                params['s_F']   = float(np.sum(F_durations).astype(np.float32))
                params['n_F']     = len(F_durations)
                params['a_F']    = float(np.mean(F_durations).astype(np.float32))
                params['m_F']  = float(np.median(F_durations).astype(np.float32))
                params['%F_in_Pr']  = params['s_F'] / params['s_Pr']

            ################# G ################# 
            # Total duration of G, Number of G, Mean duration of G
            if 'G' not in listWaveforms:
                params['s_G']   = 0 
                params['a_G']    = 0 
                params['n_G']     = 0 
                params['m_G']  = 0 
                params['%G_in_Pr']  = 0 
                # for i in range(1, int(recTime) + 1):
                #     params[f'n_G_h{i}'] = 0 
                #     params[f's_G_h{i}'] = 0 
            else:
                G_durations         = [x[1] - x[0] for x in waveformIndices['G']]
                params['s_G']   = float(np.sum(G_durations).astype(np.float32))
                params['n_G']     = len(G_durations)
                params['a_G']    = float(np.mean(G_durations).astype(np.float32))
                params['m_G']  = float(np.median(G_durations).astype(np.float32))
                params['%G_in_Pr']  = params['s_G'] / params['s_Pr']
            ################# E, E1, E2 ################# 

            # # Number of sustained E2 (I dont understand what this is ) 
            # # Time from the start of E1 to the end of the EPG record (Z)
            # if 'E1' not in listWaveforms:
            #     params['t_start_E1_>end'] = 0 
            # else:
            #     params['t_start_E1_>end'] = len(recRecording)//100 - waveformIndices['E1'][0][0]

            # Total duration of E1e, Number of E1e, Mean duration of E1
            if 'E1e' not in listWaveforms: 
                params['s_E1e']  = 0 
                params['a_E1e']   = 0 
                params['n_E1e']    = 0 
                params['m_E1e'] = 0 

            else:
                E1e_durations        = [x[1] - x[0] for x in waveformIndices['E1e']]
                params['s_E1e']  = float(np.sum(E1e_durations).astype(np.float32))
                params['n_E1e']    = len(waveformIndices['E1e'])
                params['a_E1e']   = float(np.mean(E1e_durations).astype(np.float32))
                params['m_E1e'] = float(np.median(E1e_durations).astype(np.float32))

            
            # Total duration of E1, Number of E1, Mean duration of E1
            if 'E1' not in listWaveforms: 
                # E1
                params['s_E1']  = 0 
                params['a_E1']   = 0 
                params['n_E1']    = 0 
                params['m_E1'] = 0 
                params['mx_E1']    = 0 
                params['%E1_in_Pr'] = 0 
                # Single E1
                params['s_sgE1']   = 0 
                params['n_sgE1']     = 0 
                params['a_sgE1']    = 0 
                params['m_sgE1']  = 0 
                params['mx_sgE1']     = 0 
                # Fraction E1
                params['s_frE1']   = 0 
                params['n_frE1']     = 0 
                params['a_frE1']    = 0 
                params['m_frE1']  = 0 
                params['mx_frE1']     = 0    
                # E12 
                params['s_E12']   = 0 
                params['n_E12']     = 0 
                params['a_E12']    = 0 
                params['m_E12']  = 0 
                params['mx_E12']     = 0        
                params['t_1stPr>1stE12'] = 0  # Time from 1st probe to 1st E12        
                # Probe/NP to first E1
                params['n_Pr>1stE1']          = 0 
                params['n_brPr>1stE1']    = 0 
                params['n_Pr.after.1stE1']       = 0 
                params['n_brPr.after.1stE1'] = 0 
                params['t_NP>1stE1']                = 0 
                
                phloem_durations = []

            else:
                # E1
                E1_durations        = [x[1] - x[0] for x in waveformIndices['E1']]
                params['s_E1']  = float(np.sum(E1_durations).astype(np.float32))
                params['n_E1']    = len(waveformIndices['E1'])
                params['a_E1']   = float(np.mean(E1_durations).astype(np.float32))
                params['m_E1'] = float(np.median(E1_durations).astype(np.float32))
                params['mx_E1']    = float(np.max(E1_durations).astype(np.float32))
                params['%E1_in_Pr'] = params['s_E1'] / params['s_Pr']
                
                # Single E1
                tmp = recAna[recAna['label'] == 4]
                singleE1_locations = []
                for i in tmp.index:
                    if recAna.loc[i-1, 'label'] != 5 and recAna.loc[i+1, 'label'] != 5:
                        singleE1_locations.append([recAna.loc[i+1, 'time'], recAna.loc[i, 'time']])
                singleE1_durations = [x[1] - x[0] for x in singleE1_locations]
                if len(singleE1_durations) == 0:
                    params['s_sgE1']   = 0 
                    params['n_sgE1']     = 0 
                    params['a_sgE1']    = 0 
                    params['m_sgE1']  = 0 
                    params['mx_sgE1']     = 0                
                else:
                    params['s_sgE1']   = float(np.sum(singleE1_durations).astype(np.float32))
                    params['n_sgE1']     = len(singleE1_durations)
                    params['a_sgE1']    = float(np.mean(singleE1_durations).astype(np.float32))
                    params['m_sgE1']  = float(np.median(singleE1_durations).astype(np.float32))
                    params['mx_sgE1']     = float(np.max(singleE1_durations).astype(np.float32))
                
                # fraction E1
                frE1_locations = [x for x in waveformIndices['E1'] if x not in singleE1_locations]
                frE1_durations = [x[1] - x[0] for x in frE1_locations]
                if len(frE1_durations) == 0:
                    params['s_frE1']   = 0 
                    params['n_frE1']     = 0 
                    params['a_frE1']    = 0 
                    params['m_frE1']  = 0 
                    params['mx_frE1']     = 0                
                else:
                    params['s_frE1']   = float(np.sum(frE1_durations).astype(np.float32))
                    params['n_frE1']     = len(frE1_durations)
                    params['a_frE1']    = float(np.mean(frE1_durations).astype(np.float32))
                    params['m_frE1']  = float(np.median(frE1_durations).astype(np.float32))
                    params['mx_frE1']     = float(np.max(frE1_durations).astype(np.float32))      
                      
                # E12: phases with both E1 and E2
                E12_locations = get_stage(recAna, stage = 'E12')
                E12_durations = [x[1] - x[0] for x in E12_locations]
                if len(E12_durations) == 0:
                    params['s_E12']   = 0 
                    params['n_E12']     = 0 
                    params['a_E12']    = 0 
                    params['m_E12']  = 0 
                    params['mx_E12']     = 0        
                    params['t_1stPr>1stE12'] = 0  # Time from 1st probe to 1st E12        
                else:
                    params['s_E12']   = float(np.sum(E12_durations).astype(np.float32))
                    params['n_E12']     = len(E12_durations)
                    params['a_E12']    = float(np.mean(E12_durations).astype(np.float32))
                    params['m_E12']  = float(np.median(E12_durations).astype(np.float32))
                    params['mx_E12']     = float(np.max(E12_durations).astype(np.float32))
                    params['t_1stPr>1stE12'] = E12_locations[0][0] - probe_locations[0][1]
                # E1 followed by E2
                E1followedE2_locations = get_stage(recAna, stage = 'E1followedE2')
                E1followedE2_durations = [x[1] - x[0] for x in E1followedE2_locations]
                if len(E1followedE2_durations) == 0:
                    params['s_E1>E2']   = 0             
                else:
                    params['s_E1>E2']   = float(np.sum(E1followedE2_durations).astype(np.float32))

                # E1 followed by sustained E2
                E1followedsE2_locations = get_stage(recAna, stage = 'E1followedsE2')
                E1followedsE2_durations = [x[1] - x[0] for x in E1followedsE2_locations]
                if len(E1followedsE2_durations) == 0:
                    params['s_E1>sE2']   = 0             
                else:
                    params['s_E1>sE2']   = float(np.sum(E1followedsE2_durations).astype(np.float32))

                firstE1                                 = waveformIndices['E1'][0]
                probes2firstE1                          = [probe for probe in probe_locations if probe[1] < firstE1[0]]
                params['n_Pr>1stE1']          = len(probes2firstE1)
                NP2firstE                               = [NPdur for NPdur in waveformIndices['NP'] if NPdur[1] < firstE1[0]]
                params['t_NP>1stE1']                = float(np.sum(NP2firstE).astype(np.float32))

                briefprobes2firstE1                     = [probe for probe in shortprobe_locations if probe[1] < firstE1[0]]
                params['n_brPr>1stE1']    = len(briefprobes2firstE1)

                probesafterfirstE1                      = [probe for probe in shortprobe_locations if probe[0] > firstE1[1]]
                params['n_Pr.after.1stE1']       = len(probesafterfirstE1)

                briefprobesafterfirstE1                 = [probe for probe in probe_locations if probe[0] > firstE1[1] \
                                                            and probe[1] - probe[0] < 180.]
                params['n_brPr.after.1stE1'] = len(briefprobesafterfirstE1)
                
                phloem_locations = get_stage(recAna, stage = 'phloem')
                phloem_durations = [x[1] - x[0] for x in phloem_locations]
                try:
                    params['E1_index']     = params['s_E1'] / (params['s_E12'] + params['s_sgE1'])
                except Exception as e:  
                    params['E1_index'] = 0  
                try:
                    params['frE1_ratio']   = params['n_frE1'] / params['n_E12']
                except Exception as e:
                    params['frE1_ratio']   = 0 
                    error_log.append('frE1_ratio ' + str(e))
                    #print('frE1_ratio ' + str(e))
            if len(phloem_locations) == 0:
                params['t_1stphloem'] = 0 
            else:
                firstphloem = phloem_locations[0]
                params['t_1stphloem'] = firstphloem[1] - firstphloem[0]

            # Total duration of E2, Number of E2, Mean duration of E2
            if 'E2' not in listWaveforms: 
                params['s_E2']  = 0 
                params['a_E2']   = 0 
                params['n_E2']    = 0 
                params['m_E2'] = 0 
                params['mx_E2']    = 0 
                params['dur_1stE2']  = 0 
                params['a_E2_per_phloem']    = 0 
                params['%E2_in_Pr'] = 0 

                params['s_sE2']  = 0 
                params['n_sE2']    = 0 
                params['a_sE2']   = 0 
                params['m_sE2'] = 0 
                params['t_1stPr>1stsE2'] = 0


                params['n_Pr>1stE2']     = 0
                params['n_Pr>1stsE2']     = 0 
                params['n_E2>1stsE2'] = 0 
                params['n_Pr_after_1stsE2']   = 0 
                params['%E2s_in_E2']   = 0 
            else:
                E2_durations        = [x[1] - x[0] for x in waveformIndices['E2']]
                params['s_E2']  = float(np.sum(E2_durations).astype(np.float32))
                params['n_E2']    = len(waveformIndices['E2'])
                params['a_E2']   = float(np.mean(E2_durations).astype(np.float32))
                params['m_E2'] = float(np.median(E2_durations).astype(np.float32))
                params['mx_E2']    = float(np.max(E2_durations).astype(np.float32))
                params['dur_1stE2']   = E2_durations[0]
                params['%E2_in_Pr'] = params['s_E2'] / params['s_Pr']

                sustainedE2_locations         = [x for x in waveformIndices['E2'] if x[1] - x[0] > 600.]
                sustainedE2_durations = [x[1] - x[0] for x in sustainedE2_locations]
                params['s_sE2']  = float(np.sum(sustainedE2_durations).astype(np.float32))
                params['n_sE2']    = len(sustainedE2_durations)
                params['a_sE2']   = float(np.mean(sustainedE2_durations).astype(np.float32))
                params['m_sE2'] = float(np.median(sustainedE2_durations).astype(np.float32))
                params['t_1stPr>1stsE2'] = sustainedE2_locations[0][0] - probe_locations[0][1]

                firstE2                             = waveformIndices['E2'][0]
                probes2firstE2                      = [probe for probe in probe_locations if probe[1] < firstE2[0]]
                params['n_Pr>1stE2']      = len(probes2firstE2)

                firstsustainedE2                    = sustainedE2_locations[0]
                probes2firstsE2                     = [probe for probe in probe_locations if probe[1] < firstsustainedE2[0]]
                params['n_Pr>1stsE2']     = len(probes2firstsE2)

                E2_2_firstsE2                       = [E2dur for E2dur in waveformIndices['E2'] if E2dur[1] < firstsustainedE2[0]]
                params['n_E2>1stsE2']         = len(E2_2_firstsE2)

                probesafterfirstsE2                 = [probe for probe in probe_locations if probe[0] > firstsustainedE2[1]]
                params['n_Pr.after.1stsE2']          = len(probesafterfirstsE2)   

                if len(phloem_locations) == 0:
                    params['a_E2_per_phloem'] = 0 
                else:
                    params['a_E2_per_phloem'] = params['s_E2']/len(phloem_locations)

                params['%E2s_in_E2']   = params['s_sE2'] / params['s_E2']
                try:
                    params['E2_ratio']     = params['s_E2'] / params['n_E12']
                except Exception as e:
                    params['E2_ratio']     = 0 
                    error_log.append('E2_ratio ' + str(e))
                    #print('E2_ratio ' + str(e))

            params['s_E'] = params['s_E1'] + params['s_E2']

            try:
                params['%E2/C']        = params['s_E2'] / params['s_C']
            except Exception as e:
                params['%E2/C']        = 0 
                error_log.append('%E2/C ' + str(e))
                #print('%E2/C ' + str(e))
                
            if 'C' not in listWaveforms or 4 not in listWaveforms:
                params['t_1stPr>1stE1'] = 0  # Time from 1st probe to 1st E

            else:
                params['t_1stPr>1stE1'] = waveformIndices['E1'][0][0] - probe_locations[0][1]
            if 'C' not in listWaveforms or 5 not in listWaveforms:
                params['t_1stPr>1stE2'] = 0  # Time from the 1st probe to 1st E2
            else:
                params['t_1stPr>1stE2'] = waveformIndices['E2'][0][0] - probe_locations[0][1]
                            
            ############## Aggregation #################
            for key in params.keys():
                params[key] = [params[key]]
            params['name'] = [recName]
            params = pd.DataFrame(params)
            params.set_index('name', inplace = True)

            self.params.append(params)
            self.error_log[recName] = error_log
        self.params = pd.concat(self.params)
        
        if export_xlsx == True:
            with pd.ExcelWriter(f'EPGParameters_{self.dataset_name}.xlsx', engine='xlsxwriter') as writer:
                self.params.to_excel(writer, sheet_name='Overall')
            print(f'Exported to {os.getcwd()}/EPGParameters_{self.dataset_name}.xlsx')

        return self.params
        
    def make_boxplot(self, params):
        pass 

    def datasetSummary(self):

        durations = {'NP': [], 'C': [], 'E1': [], 'E2': [], 'F': [], 'G': [], 'pd': []}
        counts = {'NP': 0, 'C': 0, 'E1': 0, 'E2': 0, 'F': 0, 'G': 0, 'pd': 0}
        s_length = 0
        n = len(self.recordings)
        
        for i in range(n):
            # print(self.recordings[i])
            recAna = self.recordings[i]['ana']
            waveform_intervals = get_index(recAna)
            for waveform in waveform_intervals.keys():
                for interval in waveform_intervals[waveform]:
                    start, end = interval
                    durations[waveform].append(end - start)
                    counts[waveform] += 1
            s_length += recAna.iloc[-1]['time']
        self.durations = durations
        
        # stats = {'NP': [], 'C': [], 'E1': [], 'E2': [], 'F': [], 'G': [], 'pd': []}
        stats = []
        index_col = []
        for waveform in durations.keys():
            if len(self.durations[waveform]) > 0:
                index_col.append(waveform)
                count = counts[waveform]
                ratio = round(np.sum(durations[waveform])/s_length,3)
                mean = round(np.mean(durations[waveform]),3)
                std = round(np.std(durations[waveform]),3)
                max = round(np.max(durations[waveform]),3)
                min = round(np.min(durations[waveform]),3)
                median = round(np.median(durations[waveform]),3)
                Q1 = round(np.quantile(durations[waveform],0.25),3)
                Q3 = round(np.quantile(durations[waveform],0.75),3)

                stats.append([count, ratio, mean, std, max, min, median, Q1, Q3])
            else:
                print(f'No waveform {waveform} was found in dataset.')
        self.statistics = pd.DataFrame(stats)
        self.statistics.columns = ['count', 'ratio', 'mean', 'std', 'max', 'min', 'median', 'Q1', 'Q3']
        self.statistics.index = index_col
        return self.statistics



# ============================= Label map =============================
waveform_labels = ['NP', 'C', 'E1', 'E2', 'F', 'G', 'pd']
waveform_ids = [1, 2, 4, 5, 6, 7, 8]
phloem_label = ['non-phloem','non-phloem', 'phloem', 'phloem', 'non-phloem', 'non-phloem', 'non-phloem']
probe_label = ['NP', 'probe', 'probe', 'probe', 'probe', 'probe', 'probe']


def create_map(list1, list2):
    map = {}
    for i in range(len(list1)):
        map[list1[i]] = list2[i]
    return map

def get_stage(recAna, stage = 'probe'):
    if stage == 'probe':
        labelmap = create_map(waveform_ids, probe_label)
    elif stage == 'phloem' or stage == 'E12' or stage == 'E1followedE2' or stage == 'E1followedsE2':
        labelmap = create_map(waveform_ids, phloem_label)
    Ana = recAna.copy()
    Ana['stagelabel'] = Ana.loc[:,'label'].map(labelmap)
    stages = []
    is_at_stage = False
    substage_count = 0 
    for i in range(1, len(Ana)):
        if is_at_stage == False:
            if Ana.loc[i, 'stagelabel'] == stage:
                start_index = i
                stage_start = Ana.loc[i,'time']
                is_at_stage = True
        else:
            if Ana.loc[i, 'stagelabel'] != stage:
                end_index = i
                stage_end = Ana.loc[i,'time']
                if stage == 'E12':
                    n_substage = end_index - start_index
                    if n_substage > 1: # E12 is simply a phloem stage with more than 1 E1 and E2 components
                        stages.append([stage_start, stage_end])
                if stage == 'E1followedE2':
                    n_substage = end_index - start_index
                    if n_substage == 2: # This specific phase start with E1 and ends with the succeeding E2
                        stages.append([stage_start, stage_end])   
                    is_at_stage = False
                if stage == 'E1followedsE2': # E1 followed by sustained E2
                    n_substage = end_index - start_index
                    lenE2 = Ana.loc[end_index + 1,'time'] - Ana.loc[end_index,'time']
                    if n_substage == 2 and lenE2 > 600.: 
                        stages.append([stage_start, stage_end])                 
                else:
                    stages.append([stage_start, stage_end])
                is_at_stage = False 

    return stages 

def interval_intersection(a, b):
    '''
        Inputs: intervals, or list of intervals
    '''
    def _get_intersection(itv1, itv2):
        start1, end1 = itv1 
        start2, end2 = itv2 
        start, end = max(start1, start2), min(end1,end2)
        if end <= start:
            return None
        else:
            return [start, end]
    if isinstance(a, list) or isinstance(a[0], tuple):
        if isinstance(a[0], list) or isinstance(a[0], tuple):
            intersections = []
            for itv1 in a:
                for itv2 in b:
                    intsec = _get_intersection(itv1, itv2)
                    if intsec:
                        intersections.append(intsec)
            return intersections 
        elif isinstance(a[0], float) and isinstance(b[0], float):
            return _get_intersection(a, b)
    else:
        raise RuntimeError('Expect list or tuple type input')


import torch
from torch.utils.data import Dataset, DataLoader
from sklearn.model_selection import train_test_split
device = 'cuda' if torch.cuda.is_available() else 'cpu'
class custom_dataset(Dataset):
    def __init__(self,X,y):
        self.X = X
        self.y = y 
    def __len__(self):
        return len(self.X)
    def __getitem__(self,i):
        return self.X[i], self.y[i]
    
def train_validation_test_split(data, labels, r = [0.7, 0.1, 0.2], random_state = 28):

    ### Train/val/test split
    assert (len(r) == 3), 'Parameter r must be a list/tuple of length 3'
    r_train, r_val, r_test = r
    X, x_test, Y, y_test = train_test_split(data, labels, test_size = r_test, random_state = random_state, stratify = labels)
    x_train, x_val, y_train, y_val = train_test_split(X, Y, test_size = r_val/r_train, random_state = random_state, stratify = Y)
    
    return x_train, x_val, x_test, y_train, y_val, y_test 

def to_tensor_split(split, unsqueeze = False):
    
    x_train, x_val, x_test, y_train, y_val, y_test = split
    ### To torch.tensor 
    x_train = torch.from_numpy(x_train).float().to(device)
    x_val = torch.from_numpy(x_val).float().to(device)
    x_test = torch.from_numpy(x_test).float().to(device)

    y_train = torch.from_numpy(y_train).long().to(device)
    y_val = torch.from_numpy(y_val).long().to(device)
    y_test = torch.from_numpy(y_test).long().to(device)

    ### Reshape
    if unsqueeze == True:
        return x_train.unsqueeze(1), x_val.unsqueeze(1), x_test.unsqueeze(1), y_train, y_val, y_test
    else:
        return x_train, x_val, x_test, y_train, y_val, y_test

def get_loaders(data, labels, r = [0.7, 0.1, 0.2], batch_size = 32, random_state = 28, unsqueeze = False):
    x_train, x_val, x_test, y_train, y_val, y_test \
        = to_tensor_split( train_validation_test_split(data, labels, r = r, random_state = 28),unsqueeze )

    train_dataset = custom_dataset(x_train,y_train)
    val_dataset = custom_dataset(x_val,y_val)
    test_dataset = custom_dataset(x_test,y_test)

    train_loader = DataLoader(train_dataset, batch_size = batch_size, shuffle = True)
    val_loader = DataLoader(val_dataset, batch_size = batch_size, shuffle = False)
    test_loader = DataLoader(test_dataset, batch_size = batch_size, shuffle = False)

    print(f'Train, validate, test set sizes: {x_train.shape[0],x_val.shape[0],x_test.shape[0]}')
    print(f'Input shape:',(tuple(next(iter(train_loader))[0].shape)))

    return train_loader, val_loader, test_loader

